import numpy as np
from fspecial import fspecial_gaussian_2d

def inbounds(img, y, x):
    return 0 <= y and y < img.shape[0] and \
           0 <= x and x < img.shape[1]

def comparePatches(patch1, patch2, kernel, sigma):
    return np.exp(-np.sum(kernel*(patch1 - patch2) ** 2)/(2*sigma**2))

def nonlocalmeans(img, searchWindowRadius, averageFilterRadius, sigma, nlmSigma):
    # Initialize output to 0
    out = np.zeros_like(img)
    # Pad image to reduce boundary artifacts
    pad = max(averageFilterRadius, searchWindowRadius)
    imgPad = np.pad(img, pad)
    imgPad = imgPad[..., pad:-pad] # Don't pad color channel

    # Smoothing kernel
    filtSize = (2*averageFilterRadius + 1, 2*averageFilterRadius + 1)
    kernel = fspecial_gaussian_2d(filtSize, sigma)
    # Add third axis for broadcasting
    kernel = kernel[:, :, np.newaxis]
    for y in range(img.shape[0]):
        for x in range(img.shape[1]):
            centerPatch = imgPad[y+pad-averageFilterRadius:y+pad+averageFilterRadius+1,
                                 x+pad-averageFilterRadius:x+pad+averageFilterRadius+1,
                                 :]
            window = imgPad[y+pad-searchWindowRadius:y+pad+searchWindowRadius+1,
                            x+pad-searchWindowRadius:x+pad+searchWindowRadius+1,
                            :]
            # Go over a window around the current pixel, compute weights
            # based on difference of neighborhoods.
            weights = np.zeros((2*searchWindowRadius+1, 2*searchWindowRadius+1, 1))
            max_weight = 0
            for i, dy in enumerate(range(-searchWindowRadius, searchWindowRadius+1)):
                for j, dx in enumerate(range(-searchWindowRadius, searchWindowRadius+1)):
                    if dy == 0 and dx == 0:
                        continue # Ignore center pixel
                    elif not inbounds(img, y+dy, x+dx):
                        continue # Ignore out-of-bounds pixels
                    otherPatch = imgPad[y+dy+pad-averageFilterRadius:y+dy+pad+averageFilterRadius+1,
                                        x+dx+pad-averageFilterRadius:x+dx+pad+averageFilterRadius+1,
                                        :]
                    weight = comparePatches(centerPatch, otherPatch, kernel, nlmSigma)
                    weights[i, j, 0] = weight
                    # Trick: Use center of window to store maximum weight
                    if weight > max_weight:
                        weights[searchWindowRadius, searchWindowRadius] = weight
                        max_weight = weight

            weights /= np.sum(weights)
            out[y, x, :] = np.sum(weights * window, axis=(0,1))
    return out
